# -*- coding:utf-8 _*-  
"""
@author: Maureen Hsu
@file: dataset.py 
@time: 2020/06/02
"""

# python packages
from absl.flags import FLAGS

# 3rd-party packages
import imgaug as ia
import numpy as np
import tensorflow as tf
from loguru import logger

# self-defined packages
from utils.flag_utils import get_flag

IMAGE_FEATURE_MAP = {
    'image/width': tf.io.FixedLenFeature([], tf.int64),
    'image/height': tf.io.FixedLenFeature([], tf.int64),
    'image/filename': tf.io.FixedLenFeature([], tf.string),
    'image/source_id': tf.io.FixedLenFeature([], tf.string),
    # 'image/key/sha256': tf.io.FixedLenFeature([], tf.string),
    'image/path': tf.io.FixedLenFeature([], tf.string),
    # 'image/encoded': tf.io.FixedLenFeature([], tf.string),
    # 'image/format': tf.io.FixedLenFeature([], tf.string),
    'image/object/bbox/xmin': tf.io.VarLenFeature(tf.float32),
    'image/object/bbox/ymin': tf.io.VarLenFeature(tf.float32),
    'image/object/bbox/xmax': tf.io.VarLenFeature(tf.float32),
    'image/object/bbox/ymax': tf.io.VarLenFeature(tf.float32),
    # 'image/object/class/text': tf.io.VarLenFeature(tf.string),
    'image/object/class/label': tf.io.VarLenFeature(tf.int64),
    # 'image/object/class/label': tf.io.FixedLenFeature([], tf.int64),
    'image/object/difficult': tf.io.VarLenFeature(tf.int64),
    'image/object/truncated': tf.io.VarLenFeature(tf.int64),
    # 'image/object/view': tf.io.VarLenFeature(tf.string),
}


def get_gen(dataset, batch_size):
    _iter = iter(dataset)
    while True:
        batch_x, batch_y = [], []
        for i in range(batch_size):
            x, y = next(_iter)
            batch_x.append(x)
            batch_y.append(y)
        yield batch_x, batch_y


@logger.catch(reraise=True)
class TfDataset:
    @property
    def dataset(self):
        return self._dataset

    @dataset.setter
    def dataset(self, value):
        self._dataset = value

    def __init__(self, size, img_aug, num_classes, image_feature_map=None):
        self.img_aug = img_aug
        self.size = size
        self.num_classes = num_classes
        self.sum_num_classes = sum(num_classes)
        self.image_feature_map = image_feature_map if image_feature_map is not None else IMAGE_FEATURE_MAP
        self._dataset = None

    def dataset_map(self, map_func):
        self.dataset = self.dataset.map(map_func)

    def dataset_shuffle(self, buffer_size):
        self.dataset = self.dataset.shuffle(buffer_size)

    def dataset_batch(self, batch_size):
        self.dataset = self.dataset.batch(batch_size)
        # if self.org_dataset is not None:
        #     self.org_dataset = get_gen(self.org_dataset, batch_size)

    def prefetch_autotune(self):
        self.dataset = self.dataset.prefetch(buffer_size=tf.data.experimental.AUTOTUNE)

    @staticmethod
    def build_example(annotation, process_label_to_index):
        width = int(annotation.size[1])
        height = int(annotation.size[0])

        xmin = []
        ymin = []
        xmax = []
        ymax = []
        classes = []
        classes_text = []
        truncated = []
        # views = []
        difficult_obj = []

        for obj in annotation.objects:
            difficult = bool(int(obj.difficult))
            difficult_obj.append(int(difficult))

            xmin.append(float(obj.bbox.xmin))
            ymin.append(float(obj.bbox.ymin))
            xmax.append(float(obj.bbox.xmax))
            ymax.append(float(obj.bbox.ymax))
            # classes_text.append(str(obj.name).encode('utf8'))
            classes.append(process_label_to_index(obj.name))
            truncated.append(int(obj.truncated))
            # views.append(obj['pose'].encode('utf8'))

        example = tf.train.Example(features=tf.train.Features(feature={
            'image/height': tf.train.Feature(int64_list=tf.train.Int64List(value=[height])),
            'image/width': tf.train.Feature(int64_list=tf.train.Int64List(value=[width])),
            'image/filename': tf.train.Feature(bytes_list=tf.train.BytesList(value=[
                annotation.annotation_path.encode('utf8')])),
            'image/source_id': tf.train.Feature(bytes_list=tf.train.BytesList(value=[
                annotation.annotation_path.encode('utf8')])),
            # 'image/key/sha256': tf.train.Feature(bytes_list=tf.train.BytesList(value=[key.encode('utf8')])),
            'image/path': tf.train.Feature(bytes_list=tf.train.BytesList(value=[annotation.image_path.encode('utf8')])),
            # 'image/encoded': tf.train.Feature(bytes_list=tf.train.BytesList(value=[img_raw])),
            # 'image/format': tf.train.Feature(bytes_list=tf.train.BytesList(value=['png'.encode('utf8')])),
            'image/object/bbox/xmin': tf.train.Feature(float_list=tf.train.FloatList(value=xmin)),
            'image/object/bbox/xmax': tf.train.Feature(float_list=tf.train.FloatList(value=xmax)),
            'image/object/bbox/ymin': tf.train.Feature(float_list=tf.train.FloatList(value=ymin)),
            'image/object/bbox/ymax': tf.train.Feature(float_list=tf.train.FloatList(value=ymax)),
            # 'image/object/class/text': tf.train.Feature(bytes_list=tf.train.BytesList(value=classes_text)),
            'image/object/class/label': tf.train.Feature(
                int64_list=tf.train.Int64List(value=np.array(classes).ravel())),
            'image/object/difficult': tf.train.Feature(int64_list=tf.train.Int64List(value=difficult_obj)),
            'image/object/truncated': tf.train.Feature(int64_list=tf.train.Int64List(value=truncated)),
            # 'image/object/view': tf.train.Feature(bytes_list=tf.train.BytesList(value=views)),
        }))
        return example

    def load_voc_dataset(self, voc_set, process_label_to_index):
        examples = [self.build_example(annotation, process_label_to_index).SerializeToString() for annotation in
                    voc_set.annotations]
        dataset = tf.data.Dataset.from_tensor_slices(examples)
        return dataset

    @staticmethod
    def parse_detection_example(image_feature_map, num_category):
        def _parse(example_s):
            features = tf.io.parse_example(example_s, image_feature_map)

            # x_train
            x_train = tf.io.read_file(features["image/path"])
            x_train = tf.image.decode_png(x_train, channels=get_flag("channels", 3))

            # y_train
            labels = tf.reshape(tf.cast(tf.sparse.to_dense(features["image/object/class/label"]), tf.float32),
                                (-1, num_category))
            y_train = tf.stack([tf.sparse.to_dense(features['image/object/bbox/xmin']),
                                tf.sparse.to_dense(features['image/object/bbox/ymin']),
                                tf.sparse.to_dense(features['image/object/bbox/xmax']),
                                tf.sparse.to_dense(features['image/object/bbox/ymax']),
                                ], axis=1)
            y_train = tf.concat([y_train, labels], axis=1)
            return x_train, y_train

        return _parse

    # @staticmethod
    # def parse_classification_example(example_s, image_feature_map):
    #     features = tf.io.parse_example(example_s, image_feature_map)
    #
    #     # x_train
    #     x_train = tf.io.read_file(features["image/path"])
    #     x_train = tf.image.decode_png(x_train, channels=get_flag("channels", 3))
    #     xmin, ymin, xmax, ymax = tf.sparse.to_dense(features['image/object/bbox/xmin']), \
    #                              tf.sparse.to_dense(features['image/object/bbox/ymin']), \
    #                              tf.sparse.to_dense(features['image/object/bbox/xmax']), \
    #                              tf.sparse.to_dense(features['image/object/bbox/ymax'])
    #     x_train = x_train[ymin:ymax + 1, xmin: xmax + 1, :]
    #     y_train = tf.cast(tf.sparse.to_dense(features["image/object/class/label"]), tf.float32)
    #     return x_train, y_train

    def augment_data(self, x_train, y_train):
        image = x_train

        bbs = []
        for box in y_train:
            bbs.append(ia.BoundingBox(x1=box[0], y1=box[1], x2=box[2], y2=box[3], label=box[4:]))
        bboi = ia.BoundingBoxesOnImage(bbs, shape=image.shape)

        images, bboxes = self.img_aug.aug([image], [bboi])

        y_train = np.zeros((get_flag("yolo_max_boxes", 80), 4 + self.sum_num_classes), dtype=np.float32)
        bbs = bboxes[0].remove_out_of_image().clip_out_of_image()

        size = get_flag("size", self.size)

        for i, bbox in enumerate(bbs.bounding_boxes):
            # skip any coord < 0 or >= size bbox after augmentation
            if not 0 < bbox.center_x < size or not 0 < bbox.center_y < size:
                continue
            y_train[i][0:4] = bbox.x1 / size, bbox.y1 / size, bbox.x2 / size, bbox.y2 / size
            # convert label to one hot
            y_idx_start = 4
            for idx_cat in range(len(self.num_classes)):
                y_idx_end = y_idx_start + self.num_classes[idx_cat]
                y_train[i][y_idx_start:y_idx_end] = np.eye(self.num_classes[idx_cat])[int(bbox.label[idx_cat])]
                y_idx_start = y_idx_end

        x_train = tf.convert_to_tensor(images[0], dtype=tf.float32)
        y_train = tf.convert_to_tensor(y_train, dtype=tf.float32)

        return x_train, y_train

    @tf.function
    def tf_augment_data(self, x_train, y_train):
        x_train, y_train = tf.numpy_function(self.augment_data, [x_train, y_train], [tf.float32, tf.float32])

        x_train.set_shape((get_flag("size", self.size), get_flag("size", self.size),
                           get_flag("channels", 3)))
        y_train.set_shape((get_flag("yolo_max_boxes", 80), 4 + self.sum_num_classes))

        return x_train, y_train

    def load_image_dataset(self, image_paths):
        if not isinstance(image_paths, list):
            image_paths = [image_paths]
        dataset = tf.data.Dataset.from_tensor_slices(image_paths)
        return dataset

    def parse_image(self, filename):
        # x_train
        x_train = tf.io.read_file(filename)
        x_train = tf.image.decode_png(x_train, channels=get_flag("channels", 3))
        return x_train

    def augment_image(self, x_train):
        image = x_train
        images, _ = self.img_aug.aug([image])
        x_train = tf.convert_to_tensor(images[0], dtype=tf.float32)
        return x_train

    @tf.function
    def tf_augment_image(self, x_train):
        [x_train] = tf.numpy_function(self.augment_image, [x_train], [tf.float32])
        x_train.set_shape((get_flag("size", self.size), get_flag("size", self.size),
                           get_flag("channels", 3)))

        return x_train

    @staticmethod
    @tf.function
    def transform_targets_for_output(y_true, grid_size, anchor_idxs, sum_num_classes):
        # y_true: (N, boxes, (x1, y1, x2, y2, class, best_anchor))
        N = tf.shape(y_true)[0]

        # y_true_out: (N, grid, grid, anchors, [x, y, w, h, obj, class])
        y_true_out = tf.zeros(
            (N, grid_size, grid_size, tf.shape(anchor_idxs)[0], 4 + 1 + sum_num_classes))

        anchor_idxs = tf.cast(anchor_idxs, tf.int32)

        indexes = tf.TensorArray(tf.int32, 1, dynamic_size=True)
        updates = tf.TensorArray(tf.float32, 1, dynamic_size=True)
        idx = 0

        for i in tf.range(N):
            for j in tf.range(tf.shape(y_true)[1]):
                if tf.equal(y_true[i][j][2], 0):
                    continue
                anchor_eq = tf.equal(
                    anchor_idxs, tf.cast(y_true[i][j][-1], tf.int32))  # best anchor

                if tf.reduce_any(anchor_eq):
                    box = y_true[i][j][0:4]
                    box_xy = (y_true[i][j][0:2] + y_true[i][j][2:4]) / 2
                    # tf.print(box)

                    anchor_idx = tf.cast(tf.where(anchor_eq), tf.int32)
                    grid_xy = tf.cast(box_xy // (1 / grid_size), tf.int32)

                    # grid[y][x][anchor] = (tx, ty, bw, bh, obj(1), class)
                    # tf.print("i", grid_xy[0], "j", grid_xy[1], "k", anchor_idx[0][0], "c", tf.argmax(y_true[i][j][4:-1]))
                    indexes = indexes.write(
                        idx, [i, grid_xy[1], grid_xy[0], anchor_idx[0][0]])
                    value = tf.concat([box, [1], y_true[i][j][4:-1]], axis=0)
                    updates = updates.write(idx, value)
                    idx += 1

        if not tf.equal(idx, 0):
            # tf.print(indexes.stack())
            # tf.print(updates.stack())
            return tf.tensor_scatter_nd_update(y_true_out, indexes.stack(), updates.stack())
        else:
            return y_true_out

    def transform_targets(self, y_train, grid_factor, anchors, anchor_masks):
        y_outs = []
        grid_size = get_flag("size", self.size) // grid_factor

        # calculate anchor index for true boxes
        anchors = tf.cast(anchors, tf.float32)
        anchor_area = anchors[..., 0] * anchors[..., 1]
        box_wh = y_train[..., 2:4] - y_train[..., 0:2]
        box_wh = tf.tile(tf.expand_dims(box_wh, -2),
                         (1, 1, tf.shape(anchors)[0], 1))
        box_area = box_wh[..., 0] * box_wh[..., 1]
        intersection = tf.minimum(box_wh[..., 0], anchors[..., 0]) * \
                       tf.minimum(box_wh[..., 1], anchors[..., 1])
        iou = intersection / (box_area + anchor_area - intersection)
        anchor_idx = tf.cast(tf.argmax(iou, axis=-1), tf.float32)
        anchor_idx = tf.expand_dims(anchor_idx, axis=-1)

        y_train = tf.concat([y_train, anchor_idx], axis=-1)

        for anchor_idxs in anchor_masks:
            y_outs.append(self.transform_targets_for_output(y_train, grid_size, anchor_idxs, self.sum_num_classes))
            grid_size *= 2

        return tuple(y_outs)



# -*- coding:utf-8 _*-  
"""
@author: Maureen Hsu
@file: layers.py 
@time: 2020/05/22
"""

# python packages
from absl.flags import FLAGS

# 3rd-party packages
import tensorflow as tf
import tensorflow.keras.backend as K
from tensorflow.keras.layers import (
    ZeroPadding2D,
    Conv2D,
    BatchNormalization,
    LeakyReLU,
    Add,
    Input,
    UpSampling2D,
    Concatenate,
    Lambda
)
from tensorflow.keras.regularizers import l2
from tensorflow.keras.losses import (
    binary_crossentropy,

)
from tensorflow.keras.models import Model

# self-defined packages
from cv.yolov3.core.utils import (
    broadcast_iou,
)
from utils.flag_utils import get_flag


def DarknetConv(x, filters, size, strides=1, batch_norm=True):
    if strides == 1:
        padding = 'same'
    else:
        x = ZeroPadding2D(((1, 0), (1, 0)))(x)  # top left half-padding
        padding = 'valid'
    x = Conv2D(filters=filters, kernel_size=size,
               strides=strides, padding=padding,
               use_bias=not batch_norm, kernel_regularizer=l2(0.0005))(x)
    if batch_norm:
        x = BatchNormalization()(x)
        x = LeakyReLU(alpha=0.1)(x)
    return x


def DarknetResidual(x, filters):
    prev = x
    x = DarknetConv(x, filters // 2, 1)
    x = DarknetConv(x, filters, 3)
    x = Add()([prev, x])
    return x


def DarknetBlock(x, filters, blocks):
    x = DarknetConv(x, filters=filters, size=3, strides=2)
    for _ in range(blocks):
        x = DarknetResidual(x, filters)
    return x


def YoloConv(filters, name=None):
    def yolo_conv(x_in):
        if isinstance(x_in, tuple):
            inputs = Input(x_in[0].shape[1:]), Input(x_in[1].shape[1:])
            x, x_skip = inputs

            # concat with skip connection
            x = DarknetConv(x, filters, 1)
            x = UpSampling2D(2)(x)
            x = Concatenate()([x, x_skip])
        else:
            x = inputs = Input(x_in.shape[1:])

        x = DarknetConv(x, filters, size=1)
        x = DarknetConv(x, filters * 2, size=3)
        x = DarknetConv(x, filters, size=1)
        x = DarknetConv(x, filters * 2, size=3)
        x = DarknetConv(x, filters, size=1)
        return Model(inputs, x, name=name)(x_in)

    return yolo_conv


def YoloConvTiny(filters, name=None):
    def yolo_conv(x_in):
        if isinstance(x_in, tuple):
            inputs = Input(x_in[0].shape[1:]), Input(x_in[1].shape[1:])
            x, x_skip = inputs

            # concat with skip connection
            x = DarknetConv(x, filters, 1)
            x = UpSampling2D(2)(x)
            x = Concatenate()([x, x_skip])
        else:
            x = inputs = Input(x_in.shape[1:])
            x = DarknetConv(x, filters, 1)

        return Model(inputs, x, name=name)(x_in)

    return yolo_conv


def YoloOutput(filters, anchors, classes, name=None):
    total_class_length = sum(classes) + 5

    def yolo_output(x_in):
        x = inputs = Input(x_in.shape[1:])
        x = DarknetConv(x, filters * 2, 3)
        x = DarknetConv(x, anchors * total_class_length, 1, batch_norm=False)
        x = Lambda(lambda x: tf.reshape(x, (-1, tf.shape(x)[1], tf.shape(x)[2], anchors, total_class_length)))(x)
        return tf.keras.Model(inputs, x, name=name)(x_in)

    return yolo_output


# @tf.function
def calculate_category_loss(class_probs, classes):
    def _calculate_loss(i):
        num_category_class = tf.gather(classes, i)
        start_index = tf.cast(tf.reduce_sum(tf.slice(classes, [0], [i])), tf.int32)
        end_index = tf.cast(start_index + num_category_class, tf.int32)
        loss = tf.sigmoid(class_probs[..., start_index:end_index])
        return loss

    return _calculate_loss


def yolo_boxes(pred, anchors, classes):
    # pred: (batch_size, grid, grid, anchors, (x, y, w, h, obj, ...classes))
    grid_size = tf.shape(pred)[1]
    box_xy, box_wh, objectness, class_probs = tf.split(pred, (2, 2, 1, sum(classes)), axis=-1)

    box_xy = tf.sigmoid(box_xy)
    objectness = tf.sigmoid(objectness)
    class_probs = tf.sigmoid(class_probs)
    pred_box = tf.concat((box_xy, box_wh), axis=-1)  # original xywh for loss

    # !!! grid[x][y] == (y, x)
    grid = tf.meshgrid(tf.range(grid_size), tf.range(grid_size))
    grid = tf.expand_dims(tf.stack(grid, axis=-1), axis=2)  # [gx, gy, 1, 2]

    box_xy = (box_xy + tf.cast(grid, tf.float32)) / tf.cast(grid_size, tf.float32)
    box_wh = tf.exp(box_wh) * anchors

    box_x1y1 = box_xy - box_wh / 2
    box_x2y2 = box_xy + box_wh / 2
    bbox = tf.concat([box_x1y1, box_x2y2], axis=-1)

    return bbox, objectness, class_probs, pred_box


def yolo_nms(outputs, anchors, masks, classes):
    # boxes, conf, type
    b, c, t = [], [], []

    for o in outputs:
        b.append(tf.reshape(o[0], (tf.shape(o[0])[0], -1, tf.shape(o[0])[-1])))
        c.append(tf.reshape(o[1], (tf.shape(o[1])[0], -1, tf.shape(o[1])[-1])))
        t.append(tf.reshape(o[2], (tf.shape(o[2])[0], -1, tf.shape(o[2])[-1])))

    # shape = (batch_size, )
    bbox = tf.concat(b, axis=1)
    confidence = tf.concat(c, axis=1)
    class_probs = tf.concat(t, axis=1)

    """
    This operation performs non_max_suppression on the inputs per batch, across all classes. Prunes away boxes that 
    have high intersection-over-union (IOU) overlap with previously selected boxes. Bounding boxes are supplied as 
    [y1, x1, y2, x2], where (y1, x1) and (y2, x2) are the coordinates of any diagonal pair of box corners and the 
    coordinates can be provided as normalized (i.e., lying in the interval [0, 1]) or absolute. 
    Note that this algorithm is agnostic to where the origin is in the coordinate system. Also note that this algorithm 
    is invariant to orthogonal transformations and translations of the coordinate system; thus translating or 
    reflections of the coordinate system result in the same boxes being selected by the algorithm. The output of this 
    operation is the final boxes, scores and classes tensor returned after performing non_max_suppression.
    :parameter
    @@ boxes: [batch_size, num_boxes, q, 4] A 4-D float Tensor of shape [batch_size, num_boxes, q, 4]. If q is 1 then 
                same boxes are used for all classes otherwise, if q is equal to number of classes, 
                class-specific boxes are used.
    @@ scores: A 3-D float Tensor of shape [batch_size, num_boxes, num_classes] representing a single score 
                corresponding to each box (each row of boxes). 
    @@ max_output_size_per_class: A scalar integer Tensor representing the maximum number of boxes to be selected by 
                                  non-max suppression per class
    @@ max_total_size: A scalar representing the maximum number of boxes retained over all classes.
    @@ iou_threshold: A float representing the threshold for deciding whether boxes overlap too much with respect to IOU.
    @@ score_threshold: A float representing the threshold for deciding when to remove boxes based on score.
    @@ pad_per_class: If false, the output nmsed boxes, scores and classes are padded/clipped to max_total_size. 
                      If true, the output nmsed boxes, scores and classes are padded to be of length
                       max_size_per_class*num_classes, unless it exceeds max_total_size in which case it is clipped to 
                       max_total_size. Defaults to false.
    @@ clip_boxes: If true, the coordinates of output nmsed boxes will be clipped to [0, 1]. If false, output the box 
                   coordinates as it is. Defaults to true.
    """
    scores = confidence * class_probs
    boxes = tf.reshape(bbox, (tf.shape(bbox)[0], -1, 1, 4))
    scores = tf.reshape(scores, (tf.shape(scores)[0], -1, tf.shape(scores)[-1]))
    scores_combined = tf.ones(tf.shape(scores)[:-1])
    start_idx = 0
    for num_category_class in classes:
        end_idx = start_idx + num_category_class
        scores_combined *= tf.reduce_max(scores[..., start_idx:end_idx], axis=2)
        start_idx = end_idx
    scores_combined = tf.expand_dims(scores_combined, axis=-1)

    output_boxes, output_scores, output_classes, output_valid_detections = tf.image.combined_non_max_suppression(
        boxes=boxes,
        scores=scores_combined,  # [batch_size, num_boxes, num_classes]
        max_output_size_per_class=get_flag("yolo_max_boxes", 10),
        max_total_size=get_flag("yolo_max_boxes", 10),
        iou_threshold=get_flag("yolo_iou_threshold", 0.45),
        score_threshold=get_flag("yolo_score_threshold", 0.2)
    )

    return output_boxes, output_scores, output_classes, output_valid_detections, scores_combined


@tf.function
def multiclass_nms(pred):
    batch_boxes, batch_scores_combined, batch_scores = tf.split(pred, (4, 1, -1), axis=-1)
    batch_size = tf.shape(batch_boxes)[0]
    boxes = tf.zeros([0, 4])
    scores = tf.zeros([0, tf.shape(batch_scores)[-1]])
    nums = tf.zeros([0], dtype=tf.int32)

    def cond(i, *args):
        return tf.less(i, batch_size)

    @tf.function
    def body(i, boxes, scores, nums):
        selected_indices, selected_scores = tf.image.non_max_suppression_with_scores(
            tf.reshape(batch_boxes[i], (-1, 4)),
            tf.reshape(batch_scores_combined[i], (-1,)),
            max_output_size=get_flag("yolo_max_boxes", 10),
            iou_threshold=get_flag("yolo_iou_threshold", 0.45),
            score_threshold=get_flag("yolo_score_threshold", 0.2))
        selected_boxes = tf.gather(batch_boxes[i], selected_indices)
        selected_scores = tf.gather(batch_scores[i], selected_indices)

        boxes = tf.concat([boxes, selected_boxes], axis=0)
        scores = tf.concat([scores, selected_scores], axis=0)
        nums = tf.concat([nums, [tf.shape(selected_indices)[0]]], axis=0)

        return tf.add(i, 1), boxes, scores, nums

    i = tf.constant(0)
    i, boxes, scores, nums = tf.while_loop(cond, body, loop_vars=[i, boxes, scores, nums],
                                           shape_invariants=[i.get_shape(),
                                                             tf.TensorShape([None, 4]),
                                                             tf.TensorShape([None, None]),
                                                             tf.TensorShape([None])],
                                           back_prop=False, parallel_iterations=8)

    return boxes, scores, nums


def multiclass_nms_tensorarray(pred):
    batch_boxes, batch_scores_combined, batch_scores = tf.split(pred, (4, 1, -1), axis=-1)
    boxes = tf.TensorArray(tf.float32, size=0, dynamic_size=True, infer_shape=False, clear_after_read=True)
    scores = tf.TensorArray(tf.float32, size=0, dynamic_size=True, infer_shape=False, clear_after_read=True)
    nums = tf.TensorArray(tf.int32, size=0, dynamic_size=True, infer_shape=False, clear_after_read=True)

    # selected = tf.TensorArray(tf.float32, size=0, dynamic_size=True, infer_shape=False, clear_after_read=True)

    def cond(i, boxes, scores, nums):
        return tf.less(i, tf.shape(batch_boxes)[0])

    def body(i, boxes, scores, nums):
        selected_indices, selected_scores = tf.image.non_max_suppression_with_scores(
            tf.reshape(batch_boxes[i], (-1, 4)),
            tf.reshape(batch_scores_combined[i], (-1,)),
            max_output_size=get_flag("yolo_max_boxes", 10),
            iou_threshold=get_flag("yolo_iou_threshold", 0.45),
            score_threshold=get_flag("yolo_score_threshold", 0.2))
        selected_boxes = tf.gather(batch_boxes[i], selected_indices)
        selected_scores = tf.gather(batch_scores[i], selected_indices)

        boxes.write(i, selected_boxes)
        scores.write(i, selected_scores)
        nums.write(i, tf.shape(selected_indices)[0])

        # selected.write(i, tf.concat([boxes, scores], axis=-1))
        # selected.append([boxes, scores])
        return tf.add(i, 1), boxes, scores, nums

    i = tf.constant(0)
    i, boxes, scores, nums = tf.while_loop(cond, body, loop_vars=[i, boxes, scores, nums], back_prop=False)

    boxes = boxes.concat()
    scores = scores.concat()
    nums = nums.stack()

    return boxes, scores, nums


def stack_preds(outputs, classes):
    # boxes, conf, type
    b, c, t = [], [], []

    for o in outputs:
        b.append(tf.reshape(o[0], (tf.shape(o[0])[0], -1, tf.shape(o[0])[-1])))
        c.append(tf.reshape(o[1], (tf.shape(o[1])[0], -1, tf.shape(o[1])[-1])))
        t.append(tf.reshape(o[2], (tf.shape(o[2])[0], -1, tf.shape(o[2])[-1])))

    bbox = tf.concat(b, axis=1)
    confidence = tf.concat(c, axis=1)
    class_probs = tf.concat(t, axis=1)
    scores = confidence * class_probs
    batch_size = tf.size(bbox[:, 0, 0])
    boxes = tf.reshape(bbox, (batch_size, -1, 4))
    scores = tf.reshape(scores, (batch_size, -1, tf.shape(scores)[-1]))
    scores_combined = tf.ones(tf.shape(scores)[:-1])
    start_idx = 0
    for num_category_class in classes:
        end_idx = start_idx + num_category_class
        scores_combined *= tf.reduce_max(scores[..., start_idx:end_idx], axis=2)
        start_idx = end_idx
    scores_combined = tf.expand_dims(scores_combined, axis=-1)
    preds = tf.concat((boxes, scores_combined, scores), axis=-1)

    return preds

    # for n in tf.range(batch_size):
    #     selected_indices, selected_scores = tf.image.non_max_suppression_padded(
    #         boxes[n], scores_combined[n],
    #         max_output_size=get_flag("yolo_max_boxes", 10),
    #         iou_threshold=get_flag("yolo_iou_threshold", 0.45),
    #         score_threshold=get_flag("yolo_score_threshold", 0.2))
    #     selected_boxes = tf.gather(boxes[n], selected_indices)
    #     selected_scores = tf.gather(scores[n], selected_indices)


def weighted_binary_crossentropy(weights=1):
    def _calculate_weighted_binary_crossentropy(labels, output, from_logits):
        """Calculate weighted binary crossentropy between an output tensor and a target tensor.
        # Arguments
            target: A tensor with the same shape as `output`.
            output: A tensor.
            from_logits: Whether `output` is expected to be a logits tensor.
                By default, we consider that `output`
                encodes a probability distribution.
        # Returns
            A tensor.
        """
        # Note: tf.nn.sigmoid_cross_entropy_with_logits
        # expects logits, Keras expects probabilities.
        if not from_logits:
            # transform back to logits
            _epsilon = tf.convert_to_tensor(K.epsilon(), dtype=tf.float32)
            output = tf.clip_by_value(output, _epsilon, 1 - _epsilon)
            output = tf.math.log(output / (1 - output))

        return tf.nn.sigmoid_cross_entropy_with_logits(labels=labels, logits=output)

    def _weighted_binary_crossentropy(y_true, y_pred, from_logits=False):
        bce = _calculate_weighted_binary_crossentropy(y_true, y_pred, from_logits)
        weighted_bac = bce * tf.convert_to_tensor(weights, tf.float32)
        return K.mean(weighted_bac, axis=-1)

    return _weighted_binary_crossentropy


def calculate_loss(obj_mask, true_class, pred_class):
    def _calculate_loss(i, j):
        return obj_mask * binary_crossentropy(true_class[..., i:j],
                                              pred_class[..., i:j])

    return _calculate_loss


def YoloLoss(anchors, classes=[80], ignore_thresh=0.5, weights=[1, 1, 1, 1]):
    def yolo_loss(y_true, y_pred):
        # 1. transform all pred outputs
        # y_pred: (batch_size, grid, grid, anchors, (x, y, w, h, obj, ...cls))
        pred_box, pred_obj, pred_class, pred_xywh = yolo_boxes(y_pred, anchors, classes)
        pred_xy = pred_xywh[..., 0:2]
        pred_wh = pred_xywh[..., 2:4]

        # 2. transform all true outputs
        # y_true: (batch_size, grid, grid, anchors, (x1, y1, x2, y2, obj, cls))
        true_box, true_obj, true_class = tf.split(y_true, (4, 1, sum(classes)), axis=-1)
        true_xy = (true_box[..., 0:2] + true_box[..., 2:4]) / 2
        true_wh = true_box[..., 2:4] - true_box[..., 0:2]

        # give higher weights to small boxes
        box_loss_scale = 2 - true_wh[..., 0] * true_wh[..., 1]

        # 3. inverting the pred box equations
        grid_size = tf.shape(y_true)[1]
        grid = tf.meshgrid(tf.range(grid_size), tf.range(grid_size))
        grid = tf.expand_dims(tf.stack(grid, axis=-1), axis=2)
        true_xy = true_xy * tf.cast(grid_size, tf.float32) - \
                  tf.cast(grid, tf.float32)
        true_wh = tf.math.log(true_wh / anchors)
        true_wh = tf.where(tf.math.is_inf(true_wh), tf.zeros_like(true_wh), true_wh)

        # 4. calculate all masks
        obj_mask = tf.squeeze(true_obj, -1)
        # ignore false positive when iou is over threshold
        best_iou = tf.map_fn(
            lambda x: tf.reduce_max(broadcast_iou(x[0], tf.boolean_mask(x[1], tf.cast(x[2], tf.bool))), axis=-1),
            (pred_box, true_box, obj_mask), tf.float32)
        ignore_mask = tf.cast(best_iou < ignore_thresh, tf.float32)

        # 5. calculate all losses
        xy_loss = obj_mask * box_loss_scale * \
                  tf.reduce_sum(tf.square(true_xy - pred_xy), axis=-1)
        wh_loss = obj_mask * box_loss_scale * tf.reduce_sum(tf.square(true_wh - pred_wh), axis=-1)
        obj_loss = binary_crossentropy(true_obj, pred_obj)
        obj_loss = obj_mask * obj_loss + (1 - obj_mask) * ignore_mask * obj_loss

        class_loss = tf.zeros(tf.shape(obj_loss))
        start_idx = 0
        for num_category_class in classes:
            end_idx = start_idx + num_category_class
            class_loss += calculate_loss(obj_mask, true_class, pred_class)(start_idx, end_idx)
            start_idx = end_idx

        # 6. sum over (batch, gridx, gridy, anchors) => (batch, 1)
        xy_loss = tf.reduce_sum(xy_loss, axis=(1, 2, 3)) * float(weights[0])
        wh_loss = tf.reduce_sum(wh_loss, axis=(1, 2, 3)) * float(weights[1])
        obj_loss = tf.reduce_sum(obj_loss, axis=(1, 2, 3)) * float(weights[2])
        class_loss = tf.reduce_sum(class_loss, axis=(1, 2, 3)) * float(weights[3])

        return xy_loss + wh_loss + obj_loss + class_loss

    return yolo_loss



# -*- coding:utf-8 _*-  
"""
@author: Maureen Hsu
@file: telegram_bot.py 
@time: 2019/10/18
"""

# python packages
import requests
from typing import Any, Union, Optional


# 3rd-party packages


class TgBot:
    def __init__(self, token: Any):
        self.token = token
        self.send_photo_url = "https://api.telegram.org/bot{}/sendPhoto".format(token)

    def send_message(self, telegram_group_chat_id: Union[str, int],
                     text: Any, parse_mode: Optional[Any] = None) -> requests.Response:
        """
        Send message from telegram bot in group
        """
        request = "https://api.telegram.org/bot{}/sendMessage?chat_id={}&text={}".format(self.token,
                                                                                         telegram_group_chat_id,
                                                                                         text)
        if parse_mode is not None:
            request += "&parse_mode={}".format(parse_mode)
        r = requests.get(request)

        return r

    def send_photo(self, image_io, telegram_group_chat_id, caption=None) -> requests.Response:
        """
        Send photo from telegram bot in group
        :param image_io: files = {'photo': open(image_path), 'rb')}
        :param telegram_group_chat_id
        :param caption
        :return: response
        """
        data = {'chat_id': telegram_group_chat_id}
        if caption:
            data["caption"] = caption
        r = requests.post(self.send_photo_url, files={'photo': image_io}, data=data)
        image_io.close()
        return r



# -*- coding:utf-8 _*-  
"""
@author: Maureen Hsu
@file: misc.py 
@time: 2019/11/22
"""

# python packages

# 3rd-party packages
from loguru import logger


# self-defined packages

# global variables

# set up log & config
# logger.add(log_path, rotation="5 MB", retention="1 week") # filter=lambda record: record["extra"].get("name") == ""
@logger.catch(reraise=True)
class LabelDict(dict):
    def __init__(self, dic):
        super().__init__(dic)

    def __getitem__(self, item):
        for key, value in self.items():
            if key == item:
                return value
        return -1

    def get_key_from_value(self, value):
        for key in self.keys():
            if self[key] == value:
                return key
        return None



# -*- coding:utf-8 _*-  
"""
@author: Maureen Hsu
@file: utils.py 
@time: 2020/05/25
"""

# python packages
import tensorflow as tf

# 3rd-party packages
import numpy as np
from loguru import logger

# self-defined packages
from utils.flag_utils import get_flag


@logger.catch(reraise=True)
def broadcast_iou(box_1, box_2):
    # box_1: (..., (x1, y1, x2, y2))
    # box_2: (N, (x1, y1, x2, y2))

    # broadcast boxes
    box_1 = tf.expand_dims(box_1, -2)
    box_2 = tf.expand_dims(box_2, 0)
    # new_shape: (..., N, (x1, y1, x2, y2))
    new_shape = tf.broadcast_dynamic_shape(tf.shape(box_1), tf.shape(box_2))
    box_1 = tf.broadcast_to(box_1, new_shape)
    box_2 = tf.broadcast_to(box_2, new_shape)

    int_w = tf.maximum(tf.minimum(box_1[..., 2], box_2[..., 2]) -
                       tf.maximum(box_1[..., 0], box_2[..., 0]), 0)
    int_h = tf.maximum(tf.minimum(box_1[..., 3], box_2[..., 3]) -
                       tf.maximum(box_1[..., 1], box_2[..., 1]), 0)
    int_area = int_w * int_h
    box_1_area = (box_1[..., 2] - box_1[..., 0]) * (box_1[..., 3] - box_1[..., 1])
    box_2_area = (box_2[..., 2] - box_2[..., 0]) * (box_2[..., 3] - box_2[..., 1])
    return int_area / (box_1_area + box_2_area - int_area)


@logger.catch(reraise=True)
def freeze_all(model, frozen=True, until_layer=0):
    if isinstance(model, tf.keras.Model):
        if until_layer < 0:
            until_layer = len(model.layers) + until_layer
        for i, l in enumerate(model.layers):
            if i >= until_layer:
                break
            freeze_all(l, frozen)
    else:
        model.trainable = not frozen



@logger.catch(reraise=True)
def yolo_decode_predictions(model_output, output_images):
    boxes, scores, classes, nums = model_output

    decoded = []
    for i in range(len(output_images)):  # batch_size
        wh = np.flip(output_images[i].shape[0:2])
        objs = []
        for n in range(nums[i]):
            x1y1 = tuple((np.array(boxes[i][n][0:2]) * wh).astype(np.int32))
            x2y2 = tuple((np.array(boxes[i][n][2:4]) * wh).astype(np.int32))
            score = float(scores[i][n])
            label = int(classes[i][n])
            objs.append([x1y1 + x2y2, score, label])
        decoded.append(sorted(objs, key=lambda x: (x[0], x[1], x[-1])))
    return decoded


def decode_predictions(model_output, output_images):
    # boxes, scores, classes = model_output
    decoded = []
    for i in range(len(output_images)):  # batch_size
        wh = np.flip(output_images[i].shape[0:2])
        objs = []
        for box, scores, klasses in zip(*model_output[i]):
            x1y1 = tuple((np.array(box[0:2]) * wh).astype(np.int32))
            x2y2 = tuple((np.array(box[2:4]) * wh).astype(np.int32))
            objs.append([x1y1 + x2y2, scores, klasses])
        decoded.append(objs)
    return decoded


def rescale_pred_wh(pred_boxes, output_image_shape):
    wh = np.flip(output_image_shape[0:2])
    rescaled = np.array(pred_boxes * np.repeat(wh, 2)).astype(np.int32)
    return rescaled


def convert2_class(scores, classes):
    if isinstance(scores, list) or scores.ndim == 3:
        return [convert2_class(score, classes) for score in scores]
    scores = tf.split(scores, classes, axis=-1)
    selected_scores, selected_classes = [], []
    for c in range(len(classes)):
        selected_class = tf.argmax(scores[c], axis=-1)
        ind = tf.cast(tf.expand_dims(selected_class, 1), tf.int32)
        ran = tf.expand_dims(tf.range(tf.shape(selected_class)[0], dtype=tf.int32), 1)
        ind = tf.concat([ran, ind], axis=1)
        selected_score = tf.gather_nd(scores[c], ind)
        selected_classes.append(selected_class)
        selected_scores.append(selected_score)

    return np.array(tf.stack(selected_classes, axis=-1)), np.array(tf.stack(selected_scores, axis=-1))


def decode_nms(preds, test_image_shape=None):
    boxes, scores, nums = preds
    if test_image_shape:
        boxes = rescale_pred_wh(boxes, test_image_shape)
    boxes = tf.split(boxes, nums)
    scores = tf.split(scores, nums)

    boxes = [box.numpy() for box in boxes]
    scores = [score.numpy() for score in scores]

    return boxes, scores


def trim_zeros_2d(arr_2d, axis=1):
    mask = ~(arr_2d == 0).all(axis=axis)
    return arr_2d[mask]


def decode_labels(test_labels, output_images):
    decoded = []

    def multiply_scale(x, scale):
        return int(float(x) * scale)

    for i in range(len(output_images)):
        labels = []
        H, W = output_images[i].shape[:2]
        for obj in test_labels[i]:
            xmin, ymin, xmax, ymax = obj[:4]
            klass = np.argmax(obj[4:])
            if sum(obj[:4]) != 0:  # exclude zero padding
                labels.append([(multiply_scale(xmin, W), multiply_scale(ymin, H),
                                multiply_scale(xmax, W), multiply_scale(ymax, H)),
                               klass])
        decoded.append(sorted(labels, key=lambda x: (x[0], x[1], x[-1])))
    return decoded


def find_closet_box(pred, objs, thres=20):
    coords, score, klass = pred

    losses = []
    for i, obj in enumerate(objs):
        if klass != obj[-1]:
            continue
        loss = sum(np.abs(np.array(coords) - np.array(obj[0])))
        if loss < thres:
            losses.append([i, loss])
    if not losses:
        return None, None
    losses = sorted(losses, key=lambda x: x[1])
    return losses[0]



# -*- coding:utf-8 _*-  
"""
@author: Maureen Hsu
@file: train.py 
@time: 2020/05/22

# FLAGS.flags_into_string()
# FLAGS.append_flags_into_file('path/to/flagfile.txt')
"""

# python packages
import datetime
from absl import flags, app
from absl.flags import FLAGS

# 3rd-party packages
import numpy as np
import tensorflow as tf
from tqdm import tqdm
import pandas as pd
from loguru import logger
from tensorflow.keras.callbacks import (
    ModelCheckpoint,
    EarlyStopping,
    ReduceLROnPlateau
)

# self-defined packages
from utils import os_path, image_processing, flag_utils
from cv.yolov3.core.layers import (
    YoloLoss
)
from cv.yolov3.core.models import (
    YoloV3, YoloV3Tiny
)
from cv.yolov3.core import utils as yolo_utils
from utils.keras_utils import LogHistory, PlotHistory

flags.DEFINE_integer("size", default=416, help="size of image")
flags.DEFINE_integer('epochs', default=2, help='number of epochs')
flags.DEFINE_integer('batch_size', default=24, help='batch size')
flags.DEFINE_integer('early_stop', default=5, help='early stopping')
flags.DEFINE_float('learning_rate', default=1e-3, help='learning rate')
flags.DEFINE_integer('freeze', default=0, help='freeze layers before xth layer')

flags.DEFINE_boolean('tiny', False, 'yolov3 or yolov3-tiny')
flags.DEFINE_string('weights', None, 'path to weights file')
flags.DEFINE_enum('fit', 'eager_tf', ['fit', 'eager_fit', 'eager_tf'],
                  'fit: model.fit, '
                  'eager_fit: model.fit(run_eagerly=True), '
                  'eager_tf: custom GradientTape')
flags.DEFINE_enum('transfer', 'same',
                  ['none', 'darknet', 'no_output', 'frozen', 'fine_tune', 'same'],
                  'none: Training from scratch, '
                  'darknet: Transfer darknet, '
                  'no_output: Transfer all but output, '
                  'frozen: Transfer and freeze all, '
                  'fine_tune: Transfer all and freeze darknet only, '
                  'same: Use previous trained model')
flags.DEFINE_list('loss_weight', [1.0, 1.0, 1.0, 1.0], "[xy, wh, obj, class] loss weight")
flags.DEFINE_integer('weights_num_classes', None, 'specify num class for `weights` file if different, '
                                                  'useful in transfer learning with different number of classes')


def run_eager_fit(model, train_dataset, val_dataset, optimizer, loss):
    # Eager mode is great for debugging
    # Non eager graph mode is recommended for real training
    avg_loss = tf.keras.metrics.Mean('loss', dtype=tf.float32)
    avg_val_loss = tf.keras.metrics.Mean('val_loss', dtype=tf.float32)

    for epoch in range(1, FLAGS.epochs + 1):
        for batch, (images, labels) in enumerate(train_dataset):
            with tf.GradientTape() as tape:
                outputs = model(images, training=True)
                regularization_loss = tf.reduce_sum(model.losses)
                pred_loss = []
                for output, label, loss_fn in zip(outputs, labels, loss):
                    pred_loss.append(loss_fn(label, output))
                total_loss = tf.reduce_sum(pred_loss) + regularization_loss

            grads = tape.gradient(total_loss, model.trainable_variables)
            optimizer.apply_gradients(zip(grads, model.trainable_variables))

            logger.info(
                f"[Epoch{epoch}/{FLAGS.epochs}][Train][Batch {batch + 1}] "
                f"Loss: {total_loss.numpy()}, {list(map(lambda x: np.sum(x.numpy()), pred_loss))}")
            avg_loss.update_state(total_loss)

        for batch, (images, labels) in enumerate(val_dataset):
            outputs = model(images)
            regularization_loss = tf.reduce_sum(model.losses)
            pred_loss = []
            for output, label, loss_fn in zip(outputs, labels, loss):
                pred_loss.append(loss_fn(label, output))
            total_loss = tf.reduce_sum(pred_loss) + regularization_loss

            logger.info(
                f"[Epoch{epoch}/{FLAGS.epochs}][Val][Batch {batch}] "
                f"Loss: {total_loss.numpy()}, {list(map(lambda x: np.sum(x.numpy()), pred_loss))}")
            avg_val_loss.update_state(total_loss)

        logger.info(
            f"[Epoch {epoch}][Summary] train: {avg_loss.result().numpy()}, val: {avg_val_loss.result().numpy()}")
        model.save_weights(os_path.join(os_path.dirname(FLAGS.weights), "checkpoints",
                                        f'ep{epoch:03d}-loss{avg_loss.result().numpy():.3f}'
                                        f'-val_loss{avg_val_loss.result().numpy():.3f}.h5'), )
        avg_loss.reset_states()
        avg_val_loss.reset_states()


@logger.catch(reraise=True)
def train(model, train_dataset, val_dataset):

    flag_utils.log_flag()
    """Configure the model for transfer learning"""
    if FLAGS.transfer == 'none':
        pass  # Nothing to do
    elif FLAGS.transfer in ['darknet', 'no_output']:
        if FLAGS.tiny:
            model_pretrained = YoloV3Tiny()
        else:
            model_pretrained = YoloV3()
        model_pretrained.load_weights(FLAGS.weights, by_name=True)

        if FLAGS.transfer == 'darknet':
            model.get_layer('yolo_darknet').set_weights(
                model_pretrained.get_layer('yolo_darknet').get_weights())
            yolo_utils.freeze_all(model.get_layer('yolo_darknet'), until_layer=FLAGS.freeze)

        elif FLAGS.transfer == 'no_output':
            for l in model.layers:
                if not l.name.startswith('yolo_output'):
                    l.set_weights(model_pretrained.get_layer(l.name).get_weights())
                    yolo_utils.freeze_all(l, until_layer=FLAGS.freeze)

    else:
        # All other transfer require matching classes
        model.load_weights(weight_path=FLAGS.weights)
        if FLAGS.transfer == 'fine_tune':
            # freeze darknet and fine tune other layers
            darknet = model.get_layer('yolo_darknet')
            yolo_utils.freeze_all(darknet, until_layer=FLAGS.freeze)
        elif FLAGS.transfer == 'frozen':
            # freeze everything
            yolo_utils.freeze_all(model, until_layer=FLAGS.freeze)

    optimizer = tf.keras.optimizers.Adam(lr=FLAGS.learning_rate)
    loss = [YoloLoss(model.anchors[mask], classes=model.num_classes, weights=FLAGS.loss_weight) for mask in
            model.anchor_masks]

    """Start training"""
    if FLAGS.fit == 'eager_tf':
        run_eager_fit(model, train_dataset, val_dataset, optimizer, loss)
    else:
        model.compile(optimizer=optimizer, loss=loss,
                      run_eagerly=(FLAGS.mode == 'eager_fit'))
        checkpoint_path = os_path.join(os_path.dirname(FLAGS.weights), "checkpoints",
                                       'ep{epoch:03d}-loss{loss:.3f}-val_loss{val_loss:.3f}.h5')
        os_path.make_dir(os_path.dirname(checkpoint_path))
        plt_his = PlotHistory()
        callbacks = [
            ReduceLROnPlateau(verbose=1),
            EarlyStopping(patience=FLAGS.early_stop, verbose=1),
            ModelCheckpoint(checkpoint_path,
                            monitor='val_loss', verbose=0, save_best_only=False,
                            save_weights_only=True, mode='auto', save_freq='epoch'),
            # TensorBoard(log_dir='../logs'),
            LogHistory(),
            plt_his
        ]
        try:
            history = model.fit(train_dataset,
                                epochs=FLAGS.epochs,
                                callbacks=callbacks,
                                validation_data=val_dataset)
        except Exception as e:
            raise e
        finally:
            plt_his.plot(save_path=os_path.join(os_path.dirname(FLAGS.weights),
                                                f"{datetime.datetime.now().strftime('%Y%m%d-%H%M')}.png"))


@logger.catch(reraise=True)
def test(model, test_dataset, voc_set, box_thres=15):
    model.load_weights(weight_path=FLAGS.weights)

    thres = box_thres
    box_col = ["x1", "y1", "x2", "y2"]
    class_names = [x.__name__ for x in model.class_names]
    true_class_names = [f"{x}_true" for x in class_names]

    # get all data
    test_image_shape = []
    test_labels = []
    for batch_test_images, batch_test_labels in test_dataset.as_numpy_iterator():
        test_image_shape = batch_test_images.shape[1:]
        test_labels.append(batch_test_labels)
    test_labels = np.concatenate(test_labels, axis=0)

    # predict on dataset
    preds = model.predict(test_dataset)
    pred_boxes, scores = yolo_utils.decode_nms(preds, test_image_shape)

    pred_labels = yolo_utils.convert2_class(scores, model.num_classes)

    true_boxes = list(yolo_utils.rescale_pred_wh(test_labels[..., :4], test_image_shape))
    true_boxes = [yolo_utils.trim_zeros_2d(x) for x in true_boxes]

    true_labels = yolo_utils.convert2_class(test_labels[..., 4:], model.num_classes)
    true_labels = [true_labels[i][0][:len(true_boxes[i])] for i in range(len(true_labels))]

    df_preds, df_trues = pd.DataFrame(), pd.DataFrame()
    for i in range(len(pred_boxes)):  # concat by sample
        df_pred_box = pd.DataFrame.from_records(pred_boxes[i], columns=box_col)
        df_true_box = pd.DataFrame.from_records(true_boxes[i], columns=box_col)

        df_pred_label = pd.DataFrame.from_records(pred_labels[i][0], columns=class_names)
        df_true_label = pd.DataFrame.from_records(true_labels[i], columns=class_names)
        df_pred_score = pd.DataFrame.from_records(pred_labels[i][1], columns=[f"{x}_score" for x in class_names])

        df_pred = pd.concat([df_pred_box, df_pred_score, df_pred_label], axis=1)
        df_pred["index"] = i
        df_pred["obj"] = range(len(df_pred))

        df_true = pd.concat([df_true_box, df_true_label], axis=1)
        df_true["index"] = i
        df_true["obj"] = range(len(df_true))

        df_preds = df_preds.append(df_pred, ignore_index=True)
        df_trues = df_trues.append(df_true, ignore_index=True)

    total_count = len(df_preds)
    total_label_count = total_count * len(class_names)

    df_preds = df_preds.groupby(by="index").apply(lambda x:
                                                  x.sort_values(["x1", "y1"])).reset_index(drop=True)

    # # add obj_index
    # df_preds = df_preds.groupby(by="index").apply(lambda x:
    #                                               pd.concat([x.reset_index(drop=True),
    #                                                          pd.DataFrame({"obj": range(len(x))})],
    #                                                         axis=1)).reset_index(drop=True)
    # df_trues = df_trues.groupby(by="index").apply(lambda x:
    #                                               pd.concat([x.reset_index(drop=True),
    #                                                          pd.DataFrame({"obj": range(len(x))})],
    #                                                         axis=1)).reset_index(drop=True)

    # expand to size of each other
    df_trues_can_repeat = df_trues[df_trues["index"].isin(df_preds["index"].unique())]
    df_preds_repeat = df_preds.groupby(by="index").apply(
        lambda x: pd.concat([x] * len(df_trues[df_trues["index"] == x["index"][0]]))).reset_index(drop=True)
    df_trues_repeat = df_trues_can_repeat.groupby(by="index").apply(
        lambda x: pd.concat([x] * len(df_preds[df_preds["index"] == x["index"][0]]))).reset_index(drop=True)

    # reset indexes

    df_trues_repeat = df_trues_repeat.groupby(by="index").apply(lambda x: x.sort_values("obj")).reset_index(drop=True)

    df_preds_repeat["diff"] = df_preds_repeat[box_col].subtract(df_trues_repeat[box_col]).abs().sum(axis=1)
    df_preds_repeat["true_idx"] = df_trues_repeat["obj"]

    df_preds = df_preds_repeat.loc[df_preds_repeat.groupby(["index", "obj"])["diff"].idxmin()]

    df_all = pd.merge(df_preds, df_trues, how="left", left_on=["true_idx", "index"], right_on=["obj", "index"],
                      suffixes=("", "_true")).reset_index(drop=True)

    # box_diff are huge (false positive)
    df_all["w"] = df_all["x2_true"] - df_all["x1_true"]
    df_all["h"] = df_all["y2_true"] - df_all["y1_true"]
    df_all["diff_rate"] = df_all["diff"] / (df_all["w"] + df_all["h"]) / 2
    false_positive_idx = df_all[df_all["diff_rate"] > thres].index
    df_all_1 = df_all.copy()

    # label mismatch
    df_all = df_all.loc[~df_all.index.isin(false_positive_idx)]
    is_label_mismatch = df_all[class_names] != df_all[true_class_names].values
    label_mismatch_idx = df_all[is_label_mismatch.any(axis=1)].index

    # miss prediction (true negative)
    miss_label_index = df_all.groupby(by="index").apply(lambda x: np.setdiff1d(np.arange(len(x)), x["obj"]))
    miss_label_index = miss_label_index[miss_label_index.apply(lambda x: len(x) > 0)]
    if len(miss_label_index) > 0:
        df_missed = pd.DataFrame(miss_label_index, columns=["obj"]).reset_index("index").explode("obj")
        df_true_label_misses = pd.merge(df_missed, df_trues, how="left")
    else:
        df_true_label_misses = pd.DataFrame()

    df_all = df_all_1

    true_negative_count = len(df_true_label_misses)
    total_box_loss = df_all["diff"].sum()
    total_false_positive_count = len(false_positive_idx)
    class_diff = {class_name: is_label_mismatch[class_name].sum() for class_name in class_names}

    logger.debug(
        f"Total miss detection count={true_negative_count}, avg tn={true_negative_count / total_count:.4f}")
    logger.debug(f"Total box loss={total_box_loss}, avg box loss={total_box_loss / total_count:.4f}")
    logger.debug(
        f"Total false positive count={total_false_positive_count}, avg fp={total_false_positive_count / total_count:.4f}")
    logger.debug(
        f"Total misclassification: {sum(class_diff.values())}, classes diff: {class_diff} "
        f"class acc: {1 - sum(class_diff.values()) / total_label_count}")
    logger.debug(
        f"Total accuracy= {1 - (sum(class_diff.values()) + total_false_positive_count + true_negative_count) / total_count}")

    def get_bbox(df_bbox):
        return list(df_bbox.values.ravel().astype('int'))

    def get_label(classes):
        if isinstance(classes, str):
            return classes
        if not classes.empty:
            classes = classes.values.ravel()
            return "".join([str(model.class_names[i].get_pair(int(classes[i]))) for i in range(len(class_names))])
        else:
            return ""

    def display(image_path, bbox, pred_classes, true_classes, title):
        image_org = image_processing.imread(image_path)
        image = model.img_aug.aug([image_org])[0][0]
        image = image_processing.draw_box(image, bbox, label=get_label(pred_classes))
        image_path = os_path.join(os_path.dirname(image_path, depth=2, full_path=False),
                                  os_path.get_filename(image_path))
        image_processing.show(image, title=f"{title} true={get_label(true_classes)}\n {image_path}")

    # display
    for idxes, title in [[false_positive_idx, "False Positive"], [label_mismatch_idx, "Mismatch"]]:
        for i in idxes:
            row = df_all.loc[i]
            dic_row = row.to_dict()
            dic_row["image_path"] = voc_set.image_paths[int(row["index"])]
            print(dic_row)
            display(voc_set.image_paths[int(row["index"])], get_bbox(row[box_col]), row[class_names],
                    row[true_class_names],
                    title)

    for i, row in df_true_label_misses.iterrows():
        display(voc_set.image_paths[int(row["index"])], get_bbox(row[box_col]), "XX", row[class_names], "True Negative")



# -*- coding:utf-8 _*-  
"""
@author: Maureen Hsu
@file: timeit.py 
@time: 2020/06/24
"""

# python packages
import timeit

# 3rd-party packages
from loguru import logger


# self-defined packages

class Timer:
    def __init__(self, name=None):
        self.name = name
        self.start_time = timeit.default_timer()
        self.end_time = None

    def start(self):
        self.start_time = timeit.default_timer()

    def stop(self):
        self.end_time = timeit.default_timer()
        logger.debug(
            f"{self.name + ' ' if self.name is not None else ''} Time elapse {self.end_time - self.start_time}s")


def time_function(func):
    def wrap(*args, **kwargs):
        timer = Timer(f"{func.__name__}")
        func(*args, **kwargs)
        timer.stop()
    return wrap



from glob import glob
import os
import shutil
from os.path import *
from typing import Optional, Any, List, Union

from loguru import logger

IMAGE_EXT = [".jpg", ".png"]


@logger.catch(reraise=True)
def copy(path1: str, path2: str) -> None:
    shutil.copy2(path1, path2)


@logger.catch(reraise=True)
def copytree(src: str, dst: str,
             symlinks: bool = False, ignore: Optional[bool] = None, overwrite: bool = True) -> None:
    for item in os.listdir(src):
        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if overwrite and os.path.exists(d):
            rm(d)

        if os.path.isdir(s):
            shutil.copytree(s, d, symlinks, ignore)
        else:
            shutil.copy2(s, d)


@logger.catch(reraise=True)
def rename(path1: str, path2: str) -> None:
    tmp_file = f"{path1}.tmp"
    shutil.copy2(path1, tmp_file)
    try:
        os.rename(path1, path2)
    except Exception as e:
        pass
    rm(tmp_file)


@logger.catch(reraise=True)
def chdir(path: str) -> None:
    logger.debug("Current path has changed to {}".format(os.path.abspath(path)))
    os.chdir(path)


@logger.catch(reraise=True)
def join(path: Any, *paths: Any) -> str:
    path = str(path)
    paths_s = []
    for i in range(len(paths)):
        paths_s.append(str(paths[i]))
    return os.path.join(path, *paths_s)


@logger.catch(reraise=True)
def check_pattern(pattern: str, filename: str, check_all: bool = False) -> bool:
    if isinstance(pattern, list):
        return check_patterns(pattern, filename, check_all)

    if (pattern is not None and pattern in filename) or pattern is None:
        pattern_check = True
    else:
        pattern_check = False

    return pattern_check


@logger.catch(reraise=True)
def check_patterns(patterns: List[str], filename: str, check_all: bool = False) -> bool:
    for pattern in patterns:
        is_contain = check_pattern(pattern, filename, False)
        if check_all:
            if not is_contain:
                return False
        elif is_contain:
            return True

    if check_all:
        return True
    else:
        return False


@logger.catch(reraise=True)
def check_extension(extension: Union[str, List[str]], filename: str) -> bool:
    if extension is None:
        return True

    if isinstance(extension, list):
        return check_extensions(extension, filename)

    if extension == os.path.splitext(filename)[-1]:
        extension_check = True
    else:
        extension_check = False

    return extension_check


@logger.catch(reraise=True)
def check_extensions(extensions: List[str], filename: str) -> bool:
    for extension in extensions:
        if check_extension(extension, filename):
            return True
    return False


@logger.catch(reraise=True)
def list_dir(_dir: str, sort=False, recursive=False, pattern=None,
             extension=None, full_path=True, abs_path=False, exclude_pattern="\x00", check_all_pattern=False,
             sort_key=None) -> List[str]:
    paths = []
    if not os.path.exists(_dir):
        return paths

    if abs_path:
        _dir = os.path.abspath(_dir)

    if not recursive:
        files = os.listdir(_dir)

        for i in range(len(files)):
            if check_pattern(pattern, files[i], check_all_pattern) and \
                    check_extension(extension, files[i]) and \
                    (exclude_pattern and not check_pattern(exclude_pattern, files[i], check_all_pattern)):
                if full_path:
                    paths.append(os.path.join(_dir, files[i]))
                else:
                    paths.append(files[i])
    else:
        for root, directories, filenames in os.walk(_dir):
            if extension == "":
                for directory in directories:
                    if check_pattern(pattern, directory, check_all_pattern) and \
                            not check_pattern(exclude_pattern, directory, check_all_pattern):
                        if full_path:
                            paths.append(os.path.join(root, directory))
            else:
                for filename in filenames:
                    if check_pattern(pattern, filename, check_all_pattern) and \
                            check_extension(extension, filename) and \
                            not check_pattern(exclude_pattern, filename, check_all_pattern):
                        if full_path:
                            paths.append(os.path.join(root, filename))
                        else:
                            paths.append(filename)

    if sort:
        if sort_key is not None:
            paths = sorted(paths, key=sort_key)
        else:
            paths = sorted(paths)

    return paths


@logger.catch(reraise=True)
def make_dir(_dir: str) -> str:
    if not os.path.exists(_dir):
        os.makedirs(_dir)
        logger.debug(f"Directory {os.path.abspath(_dir)} is created")
    return _dir


@logger.catch(reraise=True)
def dirname(path: str, depth: int = 1, full_path: bool = True) -> str:
    _dir = os.path.dirname(path)
    if depth > 1:
        for i in range(depth - 1):
            _dir = os.path.dirname(_dir)
    if not full_path:
        _dir = os.path.basename(_dir)

    return _dir


@logger.catch(reraise=True)
def str2bool(v: Union[str, bool]) -> bool:
    if isinstance(v, bool):
        return v
    if v.lower() in ('yes', 'true', 'mytest', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise Exception('Boolean value expected.')


@logger.catch(reraise=True)
def rm(full_path: str) -> None:
    if not exists(full_path):
        return
    if os.path.isfile(full_path):
        os.unlink(full_path)
    else:
        shutil.rmtree(full_path)


@logger.catch(reraise=True)
def cleanup_folder_by_time(folder_s: str, num_keep: int, logging: bool = True,
                           count_folder_instead: bool = False) -> None:
    contents = list_dir(folder_s, exclude_pattern=".gitkeep", full_path=True)
    files = []
    for content in contents:
        if count_folder_instead:
            files = contents
        else:
            if os.path.isfile(content):
                files.append(content)
            else:
                cleanup_folder_by_time(content, num_keep, logging)

    files = sorted(files, key=os.path.getctime, reverse=True)  # newest at front (0)
    for i in range(num_keep, len(files)):
        rm(files[i])
        if logging:
            logger.debug(f"Removing {files[i]}")

    # if directory is empty, then delete
    if not list_dir(folder_s, recursive=True):
        rm(folder_s)
        if logging:
            logger.debug(f"Removing {folder_s}")


@logger.catch(reraise=True)
def get_filename(path: str) -> str:
    filename = os.path.basename(path)
    filename, _ = os.path.splitext(filename)
    return filename


@logger.catch(reraise=True)
def make_sub_dirs(_dir: Any, sub_dirs: List[Any]) -> None:
    for sub_dir in sub_dirs:
        make_dir(join(_dir, sub_dir))
